
from __future__ import division

import itertools
import math
import time
import unittest

PASO = 10000

class Draw(object):
    """Draw analysis."""

    def __init__(self, players, points, baselist=None):
        # check players is ok
        if len(players) < 2:
            raise ValueError("Draws are of two or more players (got %d)"
                             % len(players))
        self._rounds = math.log(len(players)) / math.log(2)
        if self._rounds % 1:
            raise ValueError("Players not power of 2 (got %d)" % len(players))
        self._rounds = int(self._rounds)

        # check points are ok given the rounds
        if self._rounds + 1 != len(points):
            raise ValueError("Points should be given for each round (got %d "
                             "points for %d rounds)" % (len(points),
                             self._rounds))

        # all players playing should have points from before
        if baselist is not None:
            baselist = dict(baselist)
            for player in players:
                if player not in baselist:
                    raise ValueError("Player not in baselist")
        else:
            # just all the players with zero points
            baselist = dict.fromkeys(players, 0)

        self._players = players
        self._points = points
        self._baselist = baselist

        # build the base draw, all with Nones
        base = [[None]]
        for i in range(self._rounds):
            base.append(base[-1] * 2)
        self._base_draw = base[::-1]

    def fill_draw(self, data):
        """Fills the draw with already-played games."""
        # check rows
        prv = self._players
        for row in data:
            pairs = list(itertools.izip(*[iter(prv)]*2))
            if len(pairs) != len(row):
                raise ValueError("Bad len of row %s (should match %s)" %
                                 (row, len(pairs)))
            for p, r in zip(pairs, row):
                if r is not None and r not in p:
                    raise ValueError("Player %s not in prv %s" % (p, r))
            prv = row

        # create a base draw
        self._base_draw[0] = self._players
        for i, row in enumerate(data):
            self._base_draw[i+1] = row

        base_copy = self._base_draw[:]
        for i, row in enumerate(base_copy[:]):
            if None in row:
                # not filled, we're done
                break

            to_score = base_copy[i]

            self._players = row
            del self._base_draw[0]
            points = self._points.pop(0)

            # give the points to each player
            for player in to_score:
                self._baselist[player] += points

            # subtract the score from the rest
            self._points = [p-points for p in self._points]

    def calculate(self):
        """Calculate the possibilities of each player."""
        total_iterations = 2 ** len(self._base_draw[0])
        print "Total iterations: %d" % total_iterations
        positions = dict((x, {}) for x in self._baselist)
        total_scores = 0
        step = 0
        for scorelist in self._get_score_lists(self._baselist.items()):
            total_scores += 1
            step += 1
            if step == PASO:
                print "%d k (%.3f)" % (total_scores / 1000,
                                       total_scores / total_iterations)
                step = 0
            for pos, player in enumerate(x[1] for x in scorelist):
                pos += 1
                player_pos = positions[player]
                player_pos[pos] = player_pos.get(pos, 0) + 1

        # adjust the percentage of times each position happened
        for player, player_pos in positions.items():
            for pos, cant in player_pos.items():
                player_pos[pos] = cant / total_scores

        return positions


    def _get_score_lists(self, baselist):
        """Return the different score lists."""
        for draw in self._simulate():
            new_scores = []
            for bplay, bpnts in baselist:
                new_scores.append((bpnts + draw.get(bplay, 0), bplay))
            yield sorted(new_scores, reverse=True)

    def _simulate(self):
        """Simulates the draw completion."""

        def _generate_draws(rn, players):
            """Generates the different draws."""
            pairs = itertools.izip(*[iter(players)]*2)
            base = self._base_draw[rn]
            filtered_pairs = [(b if b is not None and b in p else p)
                              for p, b in itertools.izip(pairs, base)]
            for winners in itertools.product(*filtered_pairs):
                if len(winners) > 1:
                    subdraws = _generate_draws(rn + 1, winners)
                else:
                    subdraws = [[winners]]
                for subdraw in subdraws:
                    yield [players] + subdraw

        def _assign_points(draw):
            """Assign points to players in draws."""
            # discard first row, as it has assigned points
            draw = draw[1:]

            # we traverse the draw from the champion, assigning once a
            # score per user (as we go from the champion, it gets the
            # biggest one)
            given = {}
            for points, players in zip(self._points, draw)[::-1]:
                for player in players:
                    if player not in given:
                        given[player] = points
            return given

        if self._base_draw == []:
            # already completed!
            yield self._baselist
        else:
            gen = _generate_draws(0, self._players)
            for draw in gen:
                yield _assign_points(draw)




class TestLoadDraw(unittest.TestCase):
    """Tests for Draw getting stuff to work on."""

    def test_load_empty(self):
        """Draws can not be empty."""
        self.assertRaises(ValueError, Draw, [], [])

    def test_load_badplayers_one(self):
        """Players should be more than one."""
        self.assertRaises(ValueError, Draw, ['a'], [])

    def test_load_badplayers_notpower2(self):
        """Players should be 2**n."""
        self.assertRaises(ValueError, Draw, 'abc', [])

    def test_load_badpoints_less(self):
        """Less points that it should."""
        self.assertRaises(ValueError, Draw, 'abcdefgh', '123')

    def test_load_badpoints_more(self):
        """More points that it should."""
        self.assertRaises(ValueError, Draw, 'abcdefgh', '12345')

    def test_load_badpoints_notints(self):
        """Points are integers."""
        self.assertRaises(ValueError, Draw, 'abcdefgh', '1b345')

    def test_load_ok(self):
        """Data is useful to work on."""
        players = 'abcdefgh'
        points = '1234'
        d = Draw(players, points)
        self.assertEqual(d._players, players)
        self.assertEqual(d._points, points)
        self.assertEqual(d._rounds, 3)


class TestSimulationClean(unittest.TestCase):
    """Tests for simulation result when not-played draw.

    Note that this delivers only the delta points for the delta draw.
    """

    def test_simul_2(self):
        """Simulate for 2 players."""
        d = Draw(tuple('ab'), [1, 5])
        d.fill_draw([])
        result = sorted(list(d._simulate()))
        self.assertEqual(len(result), 2)

        shouldbe = [
            {'a': 4},
            {'b': 4},
        ]

        for res in result:
            self.assertTrue(res in shouldbe, "not in shouldbe: %s" % res)
            shouldbe.remove(res)

        self.assertEqual(shouldbe, [])

    def test_simul_4(self):
        """Simulate for 4 players.

          1 2                 1 2
        case 1              case 5
        a                   a
        b b                 b a
        c                   c
        d d d               d d d

        case 2              case 6
        a                   a
        b b                 b a
        c                   c
        d c c               d c c

        case 3              case 7
        a                   a
        b b                 b a
        c                   c
        d d b               d d a

        case 4              case 8
        a                   a
        b b                 b a
        c                   c
        d c b               d c a
        """
        d = Draw(tuple('abcd'), [0, 1, 2])
        d.fill_draw([])
        result = sorted(list(d._simulate()))
        self.assertEqual(len(result), 8)

        shouldbe = [
            {        'b': 1,         'd': 2},  # case 1
            {        'b': 1, 'c': 2,       },  # case 2
            {        'b': 2,         'd': 1},  # case 3
            {        'b': 2, 'c': 1,       },  # case 4
            {'a': 1,                 'd': 2},  # case 5
            {'a': 1,         'c': 2,       },  # case 6
            {'a': 2,                 'd': 1},  # case 7
            {'a': 2,         'c': 1,       },  # case 8
        ]

        for res in result:
            self.assertTrue(res in shouldbe, "not in shouldbe: %s" % res)
            shouldbe.remove(res)

        self.assertEqual(shouldbe, [])

    def test_simul_8(self):
        """Simulate for 8 players.

        I will not build all the cases, so just taking some measures.
        """
        d = Draw(tuple('abcdefgh'), [0, 1, 2, 4])
        d.fill_draw([])
        results = sorted(list(d._simulate()))

        # how many results
        self.assertEqual(len(results), 128)

        # each result set will have the following points:
        #  4 with 0  (didn't win any match)
        #  2 with 1
        #  1 with 2
        #  1 with 4
        #  total: 8
        for res in results:
            points = sum(res.values())
            self.assertEqual(points, 8, "Points are not 8: %r" % res)

        # all players should get the same total score after all simulations run
        total = {}
        for res in results:
            for player, points in res.items():
                total[player] = total.get(player, 0) + points
        self.assertEqual(len(set(total.itervalues())), 1)


class TestSimulationPartial(unittest.TestCase):
    """Tests for simulation result with played draw.

    Note that this delivers only the delta points for the delta draw.
    """

    def test_bad_first_row_simple(self):
        """Bad value, small values."""
        d = Draw(tuple('ab'), [1, 2])
        self.assertRaises(ValueError, d.fill_draw, [['c']])

    def test_bad_first_row_complex(self):
        """Bad value, more values."""
        d = Draw(tuple('abcdefgh'), [1, 2, 3, 4])
        self.assertRaises(ValueError, d.fill_draw, ['acej'])

    def test_bad_value_second_row(self):
        """Bad value in the second row."""
        d = Draw(tuple('abcdefgh'), [1, 2, 3, 4])
        self.assertRaises(ValueError, d.fill_draw, ['acej', 'ah'])

    def test_bad_quant_second_row(self):
        """Bad value in the second row."""
        d = Draw(tuple('abcdefgh'), [1, 2, 3, 4])
        self.assertRaises(ValueError, d.fill_draw, ['aceh', 'a'])

    def test_simul_2a(self):
        """Simulate for 2 players, 'a' won."""
        d = Draw(tuple('ab'), [1, 2])
        d.fill_draw([['a']])
        result = sorted(list(d._simulate()))
        self.assertEqual(result, [{'a': 2, 'b': 1}])

    def test_simul_2b(self):
        """Simulate for 2 players, 'b' won."""
        d = Draw(tuple('ab'), [1, 2])
        d.fill_draw([['b']])
        result = sorted(list(d._simulate()))
        self.assertEqual(result, [{'b': 2, 'a': 1}])

    def test_simul_4_a(self):
        """Simulate for 4 players."""
        d = Draw(tuple('abcd'), [0, 1, 2])
        d.fill_draw([['a', None]])
        result = sorted(list(d._simulate()))
        self.assertEqual(len(result), 4)

        shouldbe = [
            {'a': 1,                 'd': 2},  # case 5
            {'a': 1,         'c': 2,       },  # case 6
            {'a': 2,                 'd': 1},  # case 7
            {'a': 2,         'c': 1,       },  # case 8
        ]

        for res in result:
            self.assertTrue(res in shouldbe, "not in shouldbe: %s" % res)
            shouldbe.remove(res)

        self.assertEqual(shouldbe, [])

    def test_simul_4_b(self):
        """Simulate for 4 players."""
        d = Draw(tuple('abcd'), [0, 1, 8])
        d.fill_draw([['b', 'c']])
        result = sorted(list(d._simulate()))
        self.assertEqual(len(result), 2)

        # it gives the delta for the last row
        shouldbe = [
            {'b': 7,},  # case 2
            {'c': 7,},  # case 4
        ]

        for res in result:
            self.assertTrue(res in shouldbe, "not in shouldbe: %s" % res)
            shouldbe.remove(res)

        self.assertEqual(shouldbe, [])

    def test_simul_4_c(self):
        """Simulate for 4 players."""
        d = Draw(tuple('abcd'), [1, 2, 3])
        d.fill_draw([['a', 'c'], ['c']])
        result = sorted(list(d._simulate()))
        self.assertEqual(result, [{'a': 2, 'b': 1, 'c': 3, 'd': 1}])


class TestCalculation(unittest.TestCase):
    """Tests that calculates the probabilities for each player."""

    def test_example_small(self):
        """Small example."""
        d = Draw(("Foo", "Bar"), [8, 16], [('Foo',  20),
                                           ('Bar', 15), ('Baz', 5)])
        res = d.calculate()
        # two results
        # Foo: 16, Bar:  8 => Foo: 36, Bar: 23, Baz: 5
        # Foo:  8, Bar: 16 => Foo: 28, Bar: 31, Baz: 5
        self.assertEqual(res, {'Foo': {1: .5, 2: .5},
                               'Bar': {1: .5, 2: .5},
                               'Baz': {3: 1},})

    def test_example_losers(self):
        """Small example with all low points."""
        d = Draw(("Foo", "Bar"), [1, 3], [('W1', 55), ('W2', 50),
                                          ('Foo',  20), ('Bar', 15)])
        res = d.calculate()
        # two results
        # Foo: 3, Bar: 1 => W1: 55, W2: 50, Foo: 23, Bar: 16
        # Foo: 1, Bar: 3 => W1: 55, W2: 50, Foo: 21, Bar: 18
        self.assertEqual(res, {'W1' : {1: 1},
                               'W2' : {2: 1},
                               'Foo': {3: 1},
                               'Bar': {4: 1},})

    def test_example_bigger_clean(self):
        """Bigger example, from scratch."""
        previous = [('a',  10), ('b', 15), ('c', 8), ('d', 12), ('e', 13)]
        d = Draw(tuple('abcd'), [2, 4, 8], previous)
        d.fill_draw([])
        res = d.calculate()
        # several results
        # a:8, b:2, c:4, d:2 => a: 18, b: 17, c: 12, d: 14, e: 13 => a b d e c
        # a:4, b:2, c:8, d:2 => a: 14, b: 17, c: 16, d: 14, e: 13 => b c d a e
        # a:2, b:8, c:4, d:2 => a: 12, b: 23, c: 12, d: 14, e: 13 => b d e c a
        # a:2, b:4, c:8, d:2 => a: 12, b: 19, c: 16, d: 14, e: 13 => b c d e a
        # a:8, b:2, c:2, d:4 => a: 18, b: 17, c: 10, d: 16, e: 13 => a b d e c
        # a:4, b:2, c:2, d:8 => a: 14, b: 17, c: 10, d: 20, e: 13 => d b a e c
        # a:2, b:8, c:2, d:4 => a: 12, b: 23, c: 10, d: 16, e: 13 => b d e a c
        # a:2, b:4, c:2, d:8 => a: 12, b: 19, c: 10, d: 20, e: 13 => d b e a c
        self.assertEqual(res, {
            'a': {1: .250,          3: .125, 4: .375, 5: .250},
            'b': {1: .500, 2: .500,                          },
            'c': {         2: .250,          4: .125, 5: .625},
            'd': {1: .250, 2: .250, 3: .500,                 },
            'e': {                  3: .375, 4: .500, 5: .125},
        })

    def test_example_bigger_partial1(self):
        """Bigger example, some played."""
        previous = [('a',  10), ('b', 15), ('c', 8), ('d', 12), ('e', 13)]
        d = Draw(tuple('abcd'), [2, 4, 8], previous)
        d.fill_draw([['a', None]])
        res = d.calculate()
        # several results
        # a:8, b:2, c:4, d:2 => a: 18, b: 17, c: 12, d: 14, e: 13 => a b d e c
        # a:4, b:2, c:8, d:2 => a: 14, b: 17, c: 16, d: 14, e: 13 => b c d a e
        # a:8, b:2, c:2, d:4 => a: 18, b: 17, c: 10, d: 16, e: 13 => a b d e c
        # a:4, b:2, c:2, d:8 => a: 14, b: 17, c: 10, d: 20, e: 13 => d b a e c
        self.assertEqual(res, {
            'a': {1: .500,          3: .250, 4: .250,        },
            'b': {1: .250, 2: .750,                          },
            'c': {         2: .250,                   5: .750},
            'd': {1: .250,          3: .750,                 },
            'e': {                           4: .750, 5: .250},
        })

    def test_example_bigger_partial2(self):
        """Bigger example, more played."""
        previous = [('a',  10), ('b', 15), ('c', 8), ('d', 12), ('e', 13)]
        d = Draw(tuple('abcd'), [2, 4, 8], previous)
        d.fill_draw([['a', 'd']])
        res = d.calculate()
        # several results
        # a:8, b:2, c:2, d:4 => a: 18, b: 17, c: 10, d: 16, e: 13 => a b d e c
        # a:4, b:2, c:2, d:8 => a: 14, b: 17, c: 10, d: 20, e: 13 => d b a e c
        self.assertEqual(res, {
            'a': {1: .5,        3: .5,             },
            'b': {       2: 1.,                    },
            'c': {                            5: 1.},
            'd': {1: .5,        3: .5,             },
            'e': {                     4: 1.,      },
        })

    def test_example_even_bigger(self):
        """Bigger example, more played."""
        previous = [('a',  10), ('b', 15), ('c', 8), ('d', 12), ('e', 13),
                    ('f', 9), ('g', 12), ('h', 7)]
        d = Draw(tuple('abcdefgh'), [2, 4, 8, 16], previous)
        d.fill_draw([['a', 'd', 'f', 'g'], ['d', 'f']])
        res = d.calculate()
        # several results
        # d won:
        #   a:4, b:2, c:2, d:16, e:2, f:8, g:4, h:2 =>
        #       a: 14, b: 17, c: 10, d: 28, e: 15, f: 17, g: 16, h: 9 =>
        #           d f b g e a c h
        # f won:
        #   a:4, b:2, c:2, d:8, e:2, f:16, g:4, h:2 =>
        #       a: 14, b: 17, c: 10, d: 20, e: 15, f: 25, g: 16, h: 9 =>
        #           f d b g e a c h
        self.assertEqual(res, {
            'a': {6: 1.},
            'b': {3: 1.},
            'c': {7: 1.},
            'd': {1: .5, 2: .5},
            'e': {5: 1.},
            'f': {1: .5, 2: .5},
            'g': {4: 1.},
            'h': {8: 1.},
        })


if __name__ == '__main__':
        d = Draw(tuple('abcdefghijklmnop'), [1, 2, 4, 8, 16], [])
        t = time.time()
        print sum(1 for x in d._simulate())
        print time.time() - t
